""" Creator of views in MySQL """
import logging
import os
import re
import secrets

from django.core.exceptions import ObjectDoesNotExist
from django.db import connection

from cropping.models import Article as CroppingArticle, Category
from articles.models import Article, Medium, ParentTag, Tag, TopMedia, User
from articles.serializers import MediumSerializer
from articles.definitions import media_regionality, meta_tags, pressespiegel

logger = logging.getLogger(__name__)


def create_view(**kwargs):
    """ Creates several MySQL-Views """
    # unused kwargs
    _ = kwargs
    logger.debug('  (re)creating partition views for Articles...')
    file_path = os.path.join(os.path.dirname(__file__), '../etc/views.sql')
    with open(file_path, encoding='UTF-8') as file:
        sql_statement = file.read()
        with connection.cursor() as cursor:
            # TODO: replace check statement with something more robust
            for statement in sql_statement.split("# ---"):
                stripped_statement = statement.replace('\n', ' ').strip()
                cursor.execute(stripped_statement)
                row = cursor.fetchone()
            logger.debug('  Done creating partition views.')
    return row


def connect_tags_with_parents(**kwargs):
    """ a pre defined categorization of the tags"""
    warnings = []
    for key, enum in meta_tags.items():
        try:
            category = Category.objects.get(name=key)
            tag, _ = Tag.objects.update_or_create(name=category.name)
            parent_tag, _ = ParentTag.objects.update_or_create(name=enum.value)
            if tag.parent.filter(name__contains=enum.value).count() > 0:
                continue
            tag.parent.all().delete()
            tag.parent.add(parent_tag)
            tag.save()
        except ObjectDoesNotExist:
            warnings.append(key)

    for key, value in pressespiegel.items():
        try:
            tag = Tag.objects.get(name=key)
            if tag.parent.filter(name__contains=value).count() > 0:
                continue
            parent_tag, _ = ParentTag.objects.update_or_create(name=value)
            tag.parent.all().delete()
            tag.parent.add(parent_tag)
            tag.save()
        except ObjectDoesNotExist:
            warnings.append(key)
    # logger.warning("Tags not found: %s", ", ".join(warnings))


def create_standard_user(**kwargs):
    """ creates user with id_tag 1 """
    if User.objects.filter(id_tag=1).count() == 0:
        user = User(id_tag=1)
        user.save()
        for tag in Tag.objects.all():
            user.favorite_tags.add(tag)
        user.save()


def assign_media_regionality(**kwargs):
    """ assigns regionality for known media outlets """
    media_outlets = Medium.objects.all()
    for medium in media_outlets:

        regionality_defined = False
        for key, enum in media_regionality.items():
            if medium.name == key:
                regionality_defined = True
                MediumSerializer().create({'name': key, 'medium_type': enum.value, 'medium_type_description': enum.value})

        if not regionality_defined:
            # .de .com .net .tv .online .org
            if re.search(r'\.de|\.com|\.net|\.tv|\.online|\.org', medium.name):
                MediumSerializer().create({'name': medium.name, 'medium_type': 'Webseite', 'medium_type_description': 'Webseite'})


def assign_media_to_cropped_articles(**kwargs):
    """
    Based on the Pressespiegel media distribution, a random medium is assigned to each generated article.
    """
    top_media = TopMedia.objects.all()
    distribution = create_distribution(top_media)
    generated_articles = CroppingArticle.objects.all()
    for article in generated_articles:
        randint = secrets.randbelow(len(distribution))
        medium_to_assign = Medium.objects.get(name=distribution[randint])
        article.medium = medium_to_assign
        article.save()


def create_distribution(top_media):
    """ Creates a distribution based on the media article count """
    distribution = []
    for medium in top_media:
        for _ in range(medium.count):
            distribution.append(medium.name)

    return distribution


def add_reach_to_medium(**kwargs):
    """ Add the reach of a mediums latest article to the mediums database """
    media = Medium.objects.all()
    for medium in media:
        articles = (
            Article.objects.exclude(source_file__name="wizard_of_oz_faker.pdf")
            .filter(medium__name=medium.name, wrongly_parsed=False)
            .order_by("-date_of_publication")
        )

        if articles and articles[0].date_of_publication > medium.last_updated:
            medium.last_updated = articles[0].date_of_publication
            if articles[0].reach:
                medium.reach = articles[0].reach
            medium.save()
