# pylint: disable=too-many-lines
""" Test class for article module """
import datetime
import logging
import json
from unittest import mock

from django.db.models.signals import post_save
from django.test import TransactionTestCase
from django.utils import timezone

from articles.models import Article, Files, Medium, MediumType, User, Tag, TopMedia, ParentTag
from articles.serializers import write_article_to_db, ArticleSerializer, CroppedArticleSerializer
from articles.definitions import MediaRegionality, MediaOverviewType
from articles.post_migrations import create_distribution, assign_media_to_cropped_articles
from articles.views import create_dates
from cropping.models import Article as CroppedArticle
from cropping.models import CropSession

logger = logging.getLogger(__name__)


class CreateArticlesWithTagsTestCase(TransactionTestCase):
    """
    This is a parent class for test cases that require an article to be in the database.
    When inherited it writes an article with two tags into the database and marks these tags as favorites for User 1.
    """

    def setUp(self):
        """
        Creates two tags 'Test1' and 'Test2' and a parent tag 'Parent1',
        adds these tags to User 1 and then writes an exemplary article 'Testartikel' to the test database.
        """
        tag_1 = Tag(name="Test1")
        tag_1.save()
        tag_2 = Tag(name="Test2")
        tag_2.save()
        parent = ParentTag(name="Parent1")
        parent.save()
        tag_1.parent.add(parent)
        tag_2.parent.add(parent)

        # add tags to User 1
        if User.objects.filter(id_tag=1).count() == 1:
            user = User.objects.get(id_tag=1)
            for tag in Tag.objects.all():
                user.favorite_tags.add(tag)
        else:
            logger.warning("No user 1 found.")

        with mock.patch('articles.signals.extract_information_from_file', autospec=True) as mocked_post_save:
            post_save.connect(mocked_post_save, sender=Files, dispatch_uid='extract_information_from_file')
            Files.objects.create(name='testfile.pdf', upload_date='18.11.2021')
            self.file = Files.objects.get(name='testfile.pdf')
            input_data = {
                'medium_name': 'MAZ',
                'medium_type': MediaRegionality.REGIONAL.value,
                'name': 'Testartikel',
                'source_file': self.file,
                'reach': 1000,
                'date_of_publication': datetime.datetime.today(),
                'text': 'Dies ist ein beispielhafter Beispieltext',
                'pages': 2,
                'wrongly_parsed': False,
                'tags': [tag_1.name, tag_2.name],
            }
            write_article_to_db(**input_data)


class TimelineListTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Timeline REST API
    defined in views.py > 'class TimelineList(APIView)'.
    """

    year = datetime.datetime.now().year
    month = datetime.datetime.now().month

    def test_get_all_timeline(self):
        """
        Tests the API without a user or tag specified in the query.
        The API should return Timeline objects for all tags ('Test1' and 'Test2').
        """
        logger.info("started TimelineListTest.test_get_standard_user_tags")

        response = self.client.get('/articles/tags/timeview?mediaOverviewType=' + MediaOverviewType.ALLTAGS.value)
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Test1',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                },
                {
                    'name': 'Test2',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                },
            ],
        )

    def test_get_standard_user_timeline(self):
        """
        Tests the API without a user specified in the query.
        The API should default to User 1 and return Timeline objects for User 1s favorite tags ('Test1' and 'Test2').
        """
        logger.info("started TimelineListTest.test_get_standard_user_tags")
        response = self.client.get('/articles/tags/timeview?mediaOverviewType=' + MediaOverviewType.USERTAGS.value)
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Test1',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                },
                {
                    'name': 'Test2',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                },
            ],
        )

    def test_get_user_timeline(self):
        """
        Sets only 'Test1' as User 1s favorite_tag and tests the API with User 1 specified in the query.
        The API should return Timeline objects for User 1s favorite tag ('Test1').
        """
        logger.info("started TimelineListTest.test_get_standard_user_tags")
        self.client.post('/articles/user/tags', {'user': 1, 'tags': ['Test1']})
        response = self.client.get(
            '/articles/tags/timeview?mediaOverviewType=' + MediaOverviewType.USERTAGS.value + '&user=1'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Test1',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                }
            ],
        )

    def test_get_query_timeline(self):
        """
        Tests the API with a tag specified in the query.
        The API should return Timeline objects for the query tag (Test2').
        """
        logger.info("started TimelineListTest.test_get_standard_user_tags")
        response = self.client.get(
            '/articles/tags/timeview?mediaOverviewType=' + MediaOverviewType.QUERYTAG.value + '&tag=Test2'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Test2',
                    'year': TimelineListTest.year,
                    'month': TimelineListTest.month,
                    'article_count': 1,
                    'reach': '1000',
                }
            ],
        )

    def test_get_with_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started TimelineListTest.test_get_with_wrong_uid")
        response = self.client.get(
            '/articles/tags/timeview?mediaOverviewType=' + MediaOverviewType.USERTAGS.value + '&user=2'
        )
        self.assertEqual(response.status_code, 400)


class OverviewTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Overview REST API
    defined in views.py > 'class OverviewList(APIView)'.
    """

    def test_create_user_app_conf(self):
        """
        Tests if the User 1 was created and whether the tags
        created in the parent class have been correctly marked as favorite_tags for User 1.
        """
        logger.info("started OverviewTest.test_create_user_app_conf")
        user = User.objects.get(id_tag=1)
        self.assertTrue(user.activated)
        self.assertEqual(user.favorite_tags.all().count(), 2)

    def test_overview_api_wo_user(self):
        """
        Tests the API without specifying a user in the query.
        The API should default to the user with id_tag = 1 and return an object for the test tags defined in the parent class.
        """
        logger.info("started OverviewTest.test_overview_api_wo_user")
        response = self.client.get('/articles/overview')
        self.assertEqual(response.status_code, 200)
        today_str = datetime.datetime.today().strftime('%Y-%m-%d')
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    "name": "Test1",
                    "article_name": "Testartikel",
                    "medium_name": "maz",
                    "date_of_publication": today_str,
                    "reach": "1000",
                },
                {
                    "name": "Test2",
                    "article_name": "Testartikel",
                    "medium_name": "maz",
                    "date_of_publication": today_str,
                    "reach": "1000",
                },
            ],
        )

    def test_overview_api_with_user(self):
        """
        Tests the API with a specific user id passed in the query.
        Creates a new user with id_tag = 2 and adds a tag 'Test2' to the users favorite_tags.
        The API should then only return an object for the tag 'Test2' (and not 'Test1', since it's not a favorite of User 2).
        """
        logger.info("started OverviewTest.test_overview_api_with_user")
        today_str = datetime.datetime.today().strftime('%Y-%m-%d')
        user = User(id_tag=2)
        user.save()
        tag2 = Tag.objects.get(name="Test2")
        user.favorite_tags.add(tag2)
        user.save()
        response = self.client.get('/articles/overview?user=' + str(user.id_tag))
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    "name": tag2.name,
                    "article_name": "Testartikel",
                    "medium_name": "maz",
                    "date_of_publication": today_str,
                    "reach": "1000",
                }
            ],
        )

    def test_overview_with_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started OverviewTest.test_overview_with_wrong_uid")
        response = self.client.get('/articles/overview?user=2')
        self.assertEqual(response.status_code, 400)


class RecentArticleTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the RecentArticlesForMedium REST API
    defined in views.py > 'class RecentArticlesForMedium(APIView)'.
    """

    def test_recent_articles_wo_user(self):
        """
        Tests the API case where neither a user nor a tag is specified in the query.
        Should return all articles from the medium 'MAZ' in the test database.
        """
        logger.info("started RecentArticleTest.test_recent_articles_wo_user")
        response = self.client.get(
            '/articles/articles/medium/recent?mediaOverviewType='
            + MediaOverviewType.ALLTAGS.value
            + '&amountOfArticles=5&medium_name=MAZ'
        )
        today_str = datetime.datetime.today().strftime('%Y-%m-%d')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Testartikel',
                    'date_of_publication': today_str,
                    'text': 'Dies ist ein beispielhafter Beispieltext',
                }
            ],
        )

    def test_recent_articles_usertags(self):
        """
        Tests the API case where only articles with the users favorite tags are taken into account.
        Creates user_2, and tag_3 and adds tag_3 to user_2's favorite tags. Writes a new article using tag_3 to the database.
        The API response should only contain the newly created article since its the only article tagged with tag_3.
        """
        logger.info("started RecentArticleTest.test_recent_articles_usertags")
        user_2 = User(id_tag=2)
        user_2.save()
        tag_3 = Tag(name="Test3")
        tag_3.save()
        user_2.favorite_tags.add(tag_3)
        user_2.save()
        today_str = datetime.datetime.today().strftime('%Y-%m-%d')

        self.file = Files.objects.get(name='testfile.pdf')
        input_data = {
            'medium_name': 'MAZ',
            'medium_type': MediaRegionality.REGIONAL.value,
            'name': 'Testartikel-2',
            'source_file': self.file,
            'reach': 1000,
            'date_of_publication': datetime.datetime.today(),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': [tag_3.name],
        }
        write_article_to_db(**input_data)

        response = self.client.get(
            '/articles/articles/medium/recent?mediaOverviewType='
            + MediaOverviewType.USERTAGS.value
            + '&amountOfArticles=5&medium_name=MAZ&user=2'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Testartikel-2',
                    'date_of_publication': today_str,
                    'text': 'Dies ist ein beispielhafter Beispieltext',
                }
            ],
        )

    def test_recent_articles_querytag(self):
        """
        Tests the API case where a specific tag ('Test1') is passed as a query.
        Should return all articles from the medium 'MAZ' with the tag 'Test1' in the test database.
        """
        logger.info("started RecentArticleTest.test_recent_articles_querytag")
        tag = Tag.objects.get(name="Test1")
        response = self.client.get(
            '/articles/articles/medium/recent?mediaOverviewType='
            + MediaOverviewType.QUERYTAG.value
            + '&amountOfArticles=5&medium_name=MAZ&tag='
            + tag.name
        )
        today_str = datetime.datetime.today().strftime('%Y-%m-%d')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'name': 'Testartikel',
                    'date_of_publication': today_str,
                    'text': 'Dies ist ein beispielhafter Beispieltext',
                }
            ],
        )


class TopMediaTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Top Media REST API
    defined in views.py > 'class TopMediaView(APIView)'.
    """

    def test_top_media_blank_api_wo_user(self):
        """
        Tests the API without a user specified in the query.
        Should return the article_count for every available medium ('MAZ') in the database.
        """
        logger.info("started TopMediaTest.test_top_media_blank_api_wo_user")
        response = self.client.get('/articles/top/media?mediaOverviewType=' + MediaOverviewType.ALLTAGS.value)
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'article_count': 1, 'medium_name': 'maz'}])

    def test_top_media_with_extra_articles(self):
        """
        Tests the API without a user specified in the query after two new articles are written to the database.
        Should return the article_count for every available medium ('MAZ', 'MOZ') in the database.
        """
        logger.info("started TopMediaTest.test_top_media_with_extra_articles")
        file = Files.objects.get(name='testfile.pdf')
        tag_1 = Tag.objects.get(name="Test1")
        tag_2 = Tag.objects.get(name="Test2")
        input_data = {
            'medium_name': 'MOZ',
            'medium_type': 'Zeitung',
            'name': 'Testartikel2',
            'source_file': file,
            'reach': 1111,
            'date_of_publication': datetime.datetime.today(),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': [tag_1.name],
        }
        write_article_to_db(**input_data)
        input_data['name'] = 'Testartikel3'
        input_data['tags'] = [tag_2.name]
        write_article_to_db(**input_data)
        response = self.client.get('/articles/top/media?mediaOverviewType=' + MediaOverviewType.ALLTAGS.value)
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [{'article_count': 2, 'medium_name': 'moz'}, {'article_count': 1, 'medium_name': 'maz'}],
        )

    def test_top_media_with_extra_tag(self):
        """
        Tests whether the API handles user.favorite_tags correctly. User 1 is specified in the query.
        Writes a new article to the database without adding its tag to User 1s favorite_tags and tests the API response.
        The new article and its medium should not be taken into account in the response.
        Then, the new articles tag is added as a favorite_tag for User 1 and the API response is tested again.
        Now, the new article and its medium should be included in the API response.
        """
        logger.info("started TopMediaTest.test_top_media_with_extra_tag")
        file = Files.objects.get(name='testfile.pdf')
        tag3_name = "Test3"
        input_data = {
            'medium_name': 'MOZ',
            'medium_type': 'Zeitung',
            'name': 'Testartikel2',
            'source_file': file,
            'reach': 1111,
            'date_of_publication': datetime.datetime.today(),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': [tag3_name],
        }
        write_article_to_db(**input_data)
        response = self.client.get(
            '/articles/top/media?mediaOverviewType=' + MediaOverviewType.USERTAGS.value + '&user=1'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'article_count': 1, 'medium_name': 'maz'}])
        user = User.objects.get(id_tag=1)
        tag3 = Tag.objects.get(name=tag3_name)
        user.favorite_tags.add(tag3)
        response = self.client.get(
            '/articles/top/media?mediaOverviewType=' + MediaOverviewType.USERTAGS.value + '&user=1'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [{'article_count': 1, 'medium_name': 'maz'}, {'article_count': 1, 'medium_name': 'moz'}],
        )

    def test_top_media_query_tag(self):
        """
        Tests the API with a specific tag ('Test1') passed as a query parameter.
        The API response should include article_counts for every medium that has articles tagged with that specific tag.
        """
        logger.info("started TopMediaTest.test_top_media_query_tag")
        response = self.client.get(
            '/articles/top/media?mediaOverviewType=' + MediaOverviewType.QUERYTAG.value + '&tag=Test1'
        )
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'article_count': 1, 'medium_name': 'maz'}])

    def test_top_media_with_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started TopMediaTest.test_top_media_with_wrong_uid")
        response = self.client.get(
            '/articles/top/media?mediaOverviewType=' + MediaOverviewType.USERTAGS.value + '&user=2'
        )
        self.assertEqual(response.status_code, 400)


class HotTopicTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Hot Topic REST API
    defined in views.py > 'class HotTopics(APIView)'.
    """

    def test_top_tag_blank(self):
        """
        Tests the API without a user specified in the query.
        The API should default to User 1 and return User 1s top tags ('Test1' and 'Test2', evenly split value of 0.5).
        """
        logger.info("started HotTopicTest.test_top_tag_blank")
        response = self.client.get('/articles/top/tags')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {
                    'article_count': 1,
                    'hotness_factor': 0.5,
                    'tag_name': 'Test1',
                    'diff_last_month': 0,
                    'diff_last_week': 0,
                },
                {
                    'article_count': 1,
                    'hotness_factor': 0.5,
                    'tag_name': 'Test2',
                    'diff_last_month': 0,
                    'diff_last_week': 0,
                },
            ],
        )

    def test_tags_net_last_week(self):
        """tests the calculation of the tag specific net article count for the last week"""
        logger.info("started HotTopicTest.test_top_netto_last_week")
        file = Files.objects.get(name='testfile.pdf')
        tag_1 = Tag.objects.get(name="Test1")
        today = datetime.datetime.today()
        input_data = {
            'medium_name': 'MAZ',
            'medium_type': 'Zeitung',
            'name': 'Testartikel3',
            'source_file': file,
            'reach': 1000,
            'date_of_publication': today - datetime.timedelta(days=today.weekday() + 1, weeks=1),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': [tag_1.name],
        }
        write_article_to_db(**input_data)
        input_data['name'] = 'Testartikel4'
        input_data['date_of_publication'] = today - datetime.timedelta(days=today.weekday() + 2)
        write_article_to_db(**input_data)
        input_data['name'] = 'Testartikel5'
        input_data['date_of_publication'] = today - datetime.timedelta(days=today.weekday() + 3)
        write_article_to_db(**input_data)
        response = self.client.get('/articles/top/tags')
        self.assertEqual(response.status_code, 200)
        res_obj = json.loads(str(response.content, encoding='utf8'))
        self.assertEqual(1, res_obj[0]['diff_last_week'])
        self.assertEqual(0, res_obj[1]['diff_last_week'])

    def test_tags_net_last_month(self):
        """tests the calculation of the tag specific net article count for the last month"""
        logger.info("started HotTopicTest.test_top_netto_last_month")
        file = Files.objects.get(name='testfile.pdf')
        tag_1 = Tag.objects.get(name="Test1")
        dates = create_dates(datetime.datetime.today())

        input_data = {
            'medium_name': 'MAZ',
            'medium_type': 'Zeitung',
            'name': 'Testartikel3',
            'source_file': file,
            'reach': 1000,
            # 3 weeks could not damage 'last week' tests if the test runs at the beginning of a month
            'date_of_publication': dates['last_month']['first_day'] + datetime.timedelta(days=10),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': [tag_1.name],
        }
        write_article_to_db(**input_data)
        input_data['date_of_publication'] = dates['before_last_month']['first_day'] + datetime.timedelta(days=20)
        input_data['name'] = 'Testartikel4'
        write_article_to_db(**input_data)
        input_data['date_of_publication'] = dates['before_last_month']['first_day'] + datetime.timedelta(days=15)
        input_data['name'] = 'Testartikel5'
        write_article_to_db(**input_data)
        response = self.client.get('/articles/top/tags')
        self.assertEqual(response.status_code, 200)
        res_obj = json.loads(str(response.content, encoding='utf8'))
        self.assertEqual(-1, res_obj[0]['diff_last_month'], res_obj[0]['tag_name'])
        self.assertEqual(0, res_obj[1]['diff_last_month'], res_obj[1]['tag_name'])

    def test_top_tag_user(self):
        """
        Tests the API with User 1 specified as a query parameter.
        The API should return User 1s top tags ('Test1' and 'Test2', evenly split value of 0.5).
        """
        logger.info("started HotTopicTest.test_top_tag_user")
        response = self.client.get('/articles/top/tags?user=1')
        self.assertEqual(response.status_code, 200)
        resp_obj = json.loads(str(response.content, encoding='utf8'))
        logger.info(resp_obj[0]['tag_name'])
        self.assertEqual('Test1', resp_obj[0]['tag_name'])
        self.assertEqual('Test2', resp_obj[1]['tag_name'])

    def test_top_tag_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started HotTopicTest.test_top_tag_wrong_uid")
        response = self.client.get('/articles/top/tags?user=2')
        self.assertEqual(response.status_code, 400)

    def test_create_dates(self):
        """
        Tests if the date creation works
        recent_time
        last_month
        before_last_month
        last_week
        before_last_week
        """
        dates = create_dates(datetime.datetime(2022, 5, 20))
        self.assertEqual(datetime.datetime(2022, 5, 20), dates['today'])
        self.assertEqual(datetime.datetime(2022, 5, 16), dates['this_monday'])
        self.assertEqual(datetime.datetime(2022, 5, 9), dates['last_monday'])
        self.assertEqual(datetime.datetime(2022, 5, 2), dates['before_last_monday'])
        self.assertEqual(datetime.datetime(2022, 4, 30), dates['last_month']['last_day'])
        self.assertEqual(datetime.datetime(2022, 4, 1), dates['last_month']['first_day'])
        self.assertEqual(datetime.datetime(2022, 3, 31), dates['before_last_month']['last_day'])
        self.assertEqual(datetime.datetime(2022, 3, 1), dates['before_last_month']['first_day'])


class UserTagsTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the User Favorite Tags REST API
    defined in views.py > 'class UserFavoriteTags(APIView)'.
    """

    def test_get_standard_user_tags(self):
        """
        Tests the API without a user specified in the query.
        The API should default to User 1 and return User 1s favorite tags ('Test1' and 'Test2').
        """
        logger.info("started UserTagsTest.test_get_standard_user_tags")
        response = self.client.get('/articles/user/tags')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'name': 'Test1'}, {'name': 'Test2'}])

    def test_get_with_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started UserTagsTest.test_get_with_wrong_uid")
        response = self.client.get('/articles/user/tags?user=2')
        self.assertEqual(response.status_code, 400)

    def test_get_new_user_tags(self):
        """
        Creates a new user and tests whether the API returns an empty array for that user.
        Then, a new tag ('Test3') is created and added to that users favorite_tags.
        The API should now return an array containing an object for tag 'Test3'.
        """
        logger.info("started UserTagsTest.test_get_new_user_tags")
        user = User(id_tag=14482)
        user.save()
        response = self.client.get('/articles/user/tags?user=14482')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [])
        tag = Tag(name="Test3")
        tag.save()
        user.favorite_tags.add(tag)
        article, _ = Article.objects.update_or_create(name='Testartikel')
        article.tags.add(tag)
        response = self.client.get('/articles/user/tags?user=14482')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'name': 'Test3'}])

    def test_post_tags(self):
        """
        Test the APIs POST method to set favorite_tags for a specific user.
        First, the GET response is verified to both contain 'Test1' and 'Test2' as User 1s favorite_tags.
        Then, by sending a POST request, only 'Test1' is set as a favorite_tag.
        To verify that POST requests success, User 1s favorite_tags are retrieved again and should now only contain 'Test1'.
        """
        logger.info("started UserTagsTest.test_post_tags")
        response = self.client.get('/articles/user/tags')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'name': 'Test1'}, {'name': 'Test2'}])
        response = self.client.post('/articles/user/tags', {'user': 1, 'tags': ['Test1']})
        self.assertEqual(response.status_code, 201)
        response = self.client.get('/articles/user/tags')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(str(response.content, encoding='utf8'), [{'name': 'Test1'}])

    def test_post_no_user(self):
        """
        Tests how the API handles a POST request without a user specified in the body.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started UserTagsTest.test_post_no_user")
        response = self.client.post('/articles/user/tags', {'tags': ['Test1']})
        self.assertEqual(response.status_code, 400)

    def test_post_no_tags(self):
        """
        Tests how the API handles a POST request without tags specified in the body.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started UserTagsTest.test_post_no_tags")
        response = self.client.post('/articles/user/tags', {'user': 1})
        self.assertEqual(response.status_code, 400)

    def test_post_no_correct_user(self):
        """
        Tests how the API handles a POST request containing a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started UserTagsTest.test_post_no_correct_user")
        response = self.client.post('/articles/user/tags', {'user': 1111, 'tags': ['Test1']})
        self.assertEqual(response.status_code, 400)


class TagsOverviewTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Tags Overview REST API
    defined in views.py > 'class TagsOverview(APIView)'.
    """

    def test_plain_api(self):
        """
        Tests whether the API correctly returns all tags and their parent tags.
        """
        logger.info("started TagsOverviewTest.test_plain_api")
        response = self.client.get('/articles/tags/overview')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [{'name': 'Parent1', 'tags': [{'name': 'Test1'}, {'name': 'Test2'}]}],
        )


class UserTagsArticleCountTest(CreateArticlesWithTagsTestCase):
    """
    This class contains test for the Article Count REST API
    defined in views.py > 'class UserTagsArticleCount(APIView)'.
    """

    def test_plain_api(self):
        """
        Tests the API without a user specified in the query. The API should default to User 1
        and return an object for User 1 favorite_tags, each containing the (total, regional, national and web-) article count per tag.
        """
        logger.info("started UserTagsArticleCountTest.test_plain_api")
        response = self.client.get('/articles/user/tags/article_count')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {"tag_name": "Test1", "count": 1, "regional_count": 1, "national_count": 0, "web_count": 0},
                {"tag_name": "Test2", "count": 1, "regional_count": 1, "national_count": 0, "web_count": 0},
            ],
        )

    def test_api_with_uid(self):
        """
        Tests the API with User 1 specified in the query.
        The API should return an object for User 1 favorite_tags,
        each containing the (total, regional, national and web-) article count per tag.
        """
        logger.info("started UserTagsArticleCountTest.test_api_with_uid")
        response = self.client.get('/articles/user/tags/article_count?mediaOverviewType=usertags&user=1')
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            str(response.content, encoding='utf8'),
            [
                {"tag_name": "Test1", "count": 1, "regional_count": 1, "national_count": 0, "web_count": 0},
                {"tag_name": "Test2", "count": 1, "regional_count": 1, "national_count": 0, "web_count": 0},
            ],
        )

    def test_api_with_wrong_uid(self):
        """
        Tests how the API handles a request with a user id that doesn't exist.
        The API should return a response with status_code 400 (Bad Request).
        """
        logger.info("started UserTagsArticleCountTest.test_api_with_wrong_uid")
        response = self.client.get('/articles/user/tags/article_count?mediaOverviewType=usertags&user=2')
        self.assertEqual(response.status_code, 400)


class ArticleTest(CreateArticlesWithTagsTestCase):
    """
    This class contains tests for the Article model defined in articles/models.py > 'class Article(models.Model)'
    and its serializer in articles/serializer.py > 'class ArticleSerializer(serializers.ModelSerializer)'.
    """

    reset_sequences = True

    def setUp(self):
        """Creates a new article in the database using Article.objects.create"""
        logger.info("started ArticleTest.setUp")
        with mock.patch('articles.signals.extract_information_from_file', autospec=True) as mocked_post_save:
            post_save.connect(mocked_post_save, sender=Files, dispatch_uid='extract_information_from_file')
            MediumType.objects.create(name='my local', description='Lokale Zeitung')
            Medium.objects.create(name='JDs Zeitung', medium_type_id=1)
            Files.objects.create(name='testfile.pdf', upload_date='18.11.2021')
            input_data = {
                'name': 'Testartikel',
                'date_of_publication': '1988-12-30',
                'reach': 1000,
                'pages': 2,
                'text': 'Dies ist ein beispielhafter Beispieltext',
                'wrongly_parsed': False,
            }
            Article.objects.create(medium_id=1, source_file_id=1, **input_data)

    def test_article_get(self):
        """
        Retrieves the newly created 'Testartikel' via Article.objects.get and
        tests whether its properties have been set and retrieved correctly.
        """
        logger.info("started ArticleTest.test_article_get")
        article = Article.objects.get(name='Testartikel')
        self.assertEqual(article.date_of_publication, datetime.date(1988, 12, 30))
        self.assertEqual(article.medium.name, 'JDs Zeitung')

    def test_serializer(self):
        """
        Creates a second test article via ArticleSerializer().create
        and tests if that article has been saved to the database by checking Article.objects.count()
        before and after creating the new article.
        """
        logger.info("started ArticleTest.test_serializer")
        self.assertEqual(Article.objects.count(), 1)
        input_data = {
            'name': 'Testartikel2',
            'date_of_publication': '2021-12-30',
            'reach': 2000,
            'pages': 3,
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'wrongly_parsed': False,
            'source_file_id': 1,
        }
        medium = Medium.objects.get(name='JDs Zeitung')
        input_data_ = {'medium': medium, **input_data}
        ArticleSerializer().create(input_data_)
        self.assertEqual(Article.objects.count(), 2)


def add_medium_helper(medium_name_list):
    """
    Receives a list of medium names.
    For every medium name in the list, writes a corresponding test article into the database.
    Inside write_article_to_db(), certain sub-mediums get merged into the main medium.

    :param medium_name_list: list of strings, e.g. ['medium_name_1', 'medium_name_2']
    """
    file = Files.objects.get(name='testfile.pdf')
    input_data = {
        'medium_type': 'Zeitung',
        'name': 'Testartikel',
        'source_file': file,
        'reach': 1000,
        'date_of_publication': '1988-12-30',
        'text': 'Dies ist ein beispielhafter Beispieltext',
        'pages': 2,
        'wrongly_parsed': False,
        'tags': ['tag1', 'tag2'],
    }

    for medium_name in medium_name_list:
        input_data['name'] = medium_name + " - Artikel"
        input_data['medium_name'] = medium_name
        write_article_to_db(**input_data)


class DbHandler(CreateArticlesWithTagsTestCase):
    """Contains test cases related to the write_article_to_db helper function"""

    reset_sequences = True

    def test_write_articles_name(self):
        """
        Retrieves the default article written to the database by CreateArticlesWithTagsTestCase
        and tests whether the name has been set correctly.
        """
        logger.info("started test_write_articles_name")
        article = Article.objects.get(name='Testartikel')
        self.assertEqual(article.reach, 1000)

    def test_write_articles_title(self):
        """
        Writes an article to the DB where the article name is passed as 'title' instead of 'name'.
        Tests whether that works as expected and that article can be retrieved by the 'name' property later on.
        """
        logger.info("started test_write_articles_title")
        input_data = {
            'medium_name': 'MAZ',
            'medium_type': 'Zeitung',
            'title': 'Testartikel1',
            'source_file': self.file,
            'reach': 1000,
            'date_of_publication': '1988-12-30',
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': ['tag1', 'tag2'],
        }
        write_article_to_db(**input_data)
        article = Article.objects.get(name='Testartikel1')
        self.assertEqual(article.reach, 1000)

    def test_write_articles_reach_string(self):
        """
        Writes an article to the DB where the article reach is passed as string instead of int.
        Tests whether that articles reach can be retrieved as a number later on.
        """
        logger.info("started test_write_articles_reach_string")
        input_data = {
            'medium_name': 'MAZ',
            'medium_type': 'Zeitung',
            'title': 'Testartikel2',
            'source_file': self.file,
            'reach': '1.000',
            'date_of_publication': '1988-12-30',
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': ['tag1', 'tag2'],
        }
        write_article_to_db(**input_data)
        article = Article.objects.get(name='Testartikel2')
        self.assertEqual(article.reach, 1000)

    def test_write_articles_reach_undefined_string(self):
        """
        Writes an article to the DB where the article reach is passed as string instead of int.
        The string is not convertible into an integer.
        Tests whether that article still gets written to the DB with a reach of 0.
        """
        logger.info("started test_write_articles_reach_string")
        input_data = {
            'medium_name': 'MAZ',
            'medium_type': 'Zeitung',
            'title': 'Testartikel3',
            'source_file': self.file,
            'reach': 'kA',
            'date_of_publication': '1988-12-30',
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': ['tag1', 'tag2'],
        }
        write_article_to_db(**input_data)
        article = Article.objects.get(name='Testartikel3')
        self.assertEqual(article.reach, 0)

    def test_write_articles_medium_with_dash(self):
        """
        Tests add_medium_helper() by passing various lists to it and checking
        whether certain sub-media are replaced by their main-medium counterpart.
        Consequently, there should be less new media in the DB than is passed to add_medium_helper()
        """
        logger.info("started test_write_articles_reach_string")
        add_medium_helper(['Märkische Allgemeine', 'Märkische Allgemeine - Potsdamer'])
        self.assertEqual(Medium.objects.filter(name='märkische allgemeine').count(), 1)

        add_medium_helper(['Lausitzer Rundschau', 'Lausitzer Rundschau Cottbus'])
        self.assertEqual(Medium.objects.filter(name='lausitzer rundschau').count(), 1)
        self.assertEqual(Medium.objects.all().count(), 3)

        add_medium_helper(['Welt Am Sonntag', 'welt.de', 'welt online', 'dw', 'die welt'])
        self.assertEqual(Medium.objects.all().count(), 6)

        add_medium_helper(['nd', 'neues deutschland', 'nd (neues deutschland)'])
        self.assertEqual(Medium.objects.all().count(), 7)

        add_medium_helper(['wochenspiegel', 'wochenspiegel potsdam'])
        self.assertEqual(Medium.objects.all().count(), 8)

        add_medium_helper(['bild', 'bild.de', 'bildplus', 'bild der frau'])
        self.assertEqual(Medium.objects.all().count(), 10)

        add_medium_helper(['Der Spiegel', 'Spiegel+'])
        self.assertEqual(Medium.objects.all().count(), 11)

        add_medium_helper(['nordkurier', 'nordkurier potsdam'])
        self.assertEqual(Medium.objects.all().count(), 12)

        add_medium_helper(['schweriner volkszeitung', 'schweriner volkszeitung - Viertel'])
        self.assertEqual(Medium.objects.all().count(), 13)

        add_medium_helper(['frankfurter rundschau', 'frankfurter rundschau - Test'])
        self.assertEqual(Medium.objects.all().count(), 14)

        add_medium_helper(['nordwest-zeitung', 'nordwest-zeitung asd'])
        self.assertEqual(Medium.objects.all().count(), 15)


class MediaAssignmentPostMigrationTest(CreateArticlesWithTagsTestCase):
    """Tests functionality related to assign_media_to_cropped_articles() in articles/post_migrations.py"""

    def test_distribution_creation(self):
        """Tests whether the media distribution is correctly created with the default medium in the database"""
        logger.info("started test_distribution_creation")
        top_media = TopMedia.objects.all()
        distribution = create_distribution(top_media)
        self.assertEqual(distribution, ['maz'])

    def test_distribution_creation_extra_medium(self):
        """Tests whether the media distribution is correctly created after adding an additional medium"""
        logger.info("started test_distribution_creation_extra_medium")
        file = Files.objects.get(name='testfile.pdf')
        input_data = {
            'medium_name': 'MOZ',
            'medium_type': 'Zeitung',
            'name': 'Testartikel2',
            'source_file': file,
            'reach': 1111,
            'date_of_publication': datetime.datetime.today(),
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
            'tags': 'Test1',
        }
        write_article_to_db(**input_data)
        top_media = TopMedia.objects.all().order_by('-count', 'name')
        distribution = create_distribution(top_media)
        self.assertEqual(distribution, ['maz', 'moz'])

    def test_media_assignment(self):
        """
        Tests the assign_media_to_cropped_articles() function.
        Therefore, a new Cropper Medium and CroppedArticle is saved to the database.
        Then, after running assign_media_to_cropped_articles(), we check whether the medium of the CroppedArticle has changed.
        """
        logger.info("started test_media_assignment")
        crop_session = CropSession(date_of_crop=timezone.now())
        crop_session.save()
        medium_type = MediumType(name='Cropper Default Type', description='Cropper Default Type')
        medium_type.save()
        default_medium = Medium.objects.create(name='Cropper Default', medium_type=medium_type)
        article, _ = CroppedArticle.objects.update_or_create(
            title='Testartikel-Cropping',
            session_id=crop_session.id,
            defaults={'text': 'Beispieltext', 'medium': default_medium},
        )
        serialized_article = CroppedArticleSerializer(article)
        self.assertEqual(
            serialized_article.data,
            {
                'title': 'Testartikel-Cropping',
                'text': 'Beispieltext',
                'categories': [],
                'session': 1,
                'medium': default_medium.id,
            },
        )
        assign_media_to_cropped_articles()
        article = CroppedArticle.objects.get(title='Testartikel-Cropping')
        serialized_article = CroppedArticleSerializer(article)
        maz = Medium.objects.get(name="MAZ")
        self.assertEqual(
            serialized_article.data,
            {'title': 'Testartikel-Cropping', 'text': 'Beispieltext', 'categories': [], 'session': 1, 'medium': maz.id},
        )


class LogUserActivityTest(TransactionTestCase):
    """
    This class contains test for the LogUserActivity REST API
    defined in views.py > 'class LogUserActivity(APIView)'.
    """

    def test_api(self):
        """Tests all defined UserActivity cases."""
        logger.info("started LogUserActivityTest.test_api")
        response = self.client.post('/articles/user/activity', {"user": "1", "userWentIdle": "true"})
        self.assertEqual(response.status_code, 200)
        response = self.client.post('/articles/user/activity', {"user": "1", "userReturned": "true"})
        self.assertEqual(response.status_code, 200)
