# pylint: disable=too-many-lines
""" Views with OpenAPI annotations """
import math
import datetime
import logging
import secrets
import os

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q, Count, Sum
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import render
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views import View
from django.views.generic import TemplateView

from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import generics
from rest_framework.response import Response
from rest_framework.views import APIView
import numpy as np
import numpy.polynomial.polynomial as poly
import pandas as pd
from plotly.offline import plot
import plotly.express as px

from one_api.settings_prod import BASE_DIR, MEDIA_ROOT
from articles.forms import FileUploadForm
from articles.models import (
    Article,
    Timeline,
    Medium,
    MediumType,
    Overview,
    TopMedia,
    Tag,
    TopTags,
    Files,
    User,
    ParentTag,
    GenFunction,
    FunctionData,
)
from articles.serializers import (
    ArticleSerializer,
    TimelineSerializer,
    MediumSerializer,
    MediumTypeSerializer,
    OverviewSerializer,
    TopMediaSerializer,
    TagSerializer,
    TopTagsSerializer,
    TagOverviewSerializer,
    RecentArticleSerializer,
    TagsRegionalInfluenceSerializer,
)
from articles.tasks import read_all_media_files, generate_by_plot, generate_by_id, generate_tag
from articles.bgr_tasks import autoremove_task
from articles.definitions import MediaRegionality, MediaOverviewType

logger = logging.getLogger(__name__)


@method_decorator(name='get', decorator=swagger_auto_schema(operation_description="Get a list of all articles"))
@method_decorator(
    name='post',
    decorator=swagger_auto_schema(
        operation_description="Insert a new article or update it (if its name is already in the db)"
    ),
)
class ArticleList(generics.ListCreateAPIView):
    """Returns an object containing all articles in the db"""

    queryset = Article.objects.all()
    serializer_class = ArticleSerializer


@method_decorator(
    name='get', decorator=swagger_auto_schema(operation_description="Get all details about an article, selected by id")
)
class ArticleDetail(generics.RetrieveUpdateDestroyAPIView):  # pylint: disable=too-many-ancestors
    """Returns all details of an article: title, text, categories, session, medium"""

    queryset = Article.objects.all()
    serializer_class = ArticleSerializer


@method_decorator(name='get', decorator=swagger_auto_schema(operation_description="Get a list of all media"))
@method_decorator(
    name='post',
    decorator=swagger_auto_schema(
        operation_description="Insert a new medium or update it (if its name is already in the db)"
    ),
)
class MediumList(generics.ListCreateAPIView):
    """Returns an object containing all media in the db"""

    queryset = Medium.objects.all()
    serializer_class = MediumSerializer


@method_decorator(
    name='get', decorator=swagger_auto_schema(operation_description="Get all details about a medium, selected by id")
)
class MediumDetail(generics.RetrieveUpdateDestroyAPIView):  # pylint: disable=too-many-ancestors
    """Returns all details of a medium: name, medium_type"""

    queryset = Medium.objects.all()
    serializer_class = MediumSerializer


@method_decorator(name='get', decorator=swagger_auto_schema(operation_description="Get a list of all media types"))
@method_decorator(
    name='post',
    decorator=swagger_auto_schema(
        operation_description="Insert a new medium_type or update it (if its name is already in the db)"
    ),
)
class MediumTypeList(generics.ListCreateAPIView):
    """Returns an object containing all medium_types in the db"""

    queryset = MediumType.objects.all()
    serializer_class = MediumTypeSerializer


@method_decorator(
    name='get',
    decorator=swagger_auto_schema(operation_description="Get all details about a medium_type, selected by id"),
)
class MediumTypeDetail(generics.RetrieveUpdateDestroyAPIView):  # pylint: disable=too-many-ancestors
    """Get details of a medium_type: name, description"""

    queryset = MediumType.objects.all()
    serializer_class = MediumTypeSerializer


def user_tags(request_params):
    """
    Get a users favorite tags, selected by user id. If no user is specified, it defaults to the user id 1

    :param request_params: an API requests query_params object
    """
    id_tag = 1
    if 'user' in request_params:
        id_tag = request_params['user']
    user = User.objects.get(id_tag=id_tag)
    tag_names = []
    for tag in user.favorite_tags.all():
        tag_names.append(tag.name)
    return tag_names


user_param_get = openapi.Parameter(
    'user', openapi.IN_QUERY, description="id which is connected to the user", type=openapi.TYPE_NUMBER
)


class TimelineList(APIView):
    """Returns Timeline objects containing monthly article_count and reach values for selected topics over the last year"""

    @swagger_auto_schema(manual_parameters=[user_param_get], responses={200: TimelineSerializer(many=True)})
    def get(self, request):
        """
        Get Timeline objects for a specified media_overview_type
            - alltags: return Timeline objects for all existing tags
            - usertag: return Timeline objects for the users favorite tags
            - querytag: return Timeline objects only for the tag specified in the query

        :query_param mediaOverviewType: defines which tags should be taken into account (alltags, usertags, querytag)
        :query_param user: id which is connected to the user (optional, only considered when mediaOverviewType=usertags, default: 1)
        :query_param tag: name of the tag to return Timeline objects for (required when mediaOverviewType=querytag)
        """
        try:
            media_overview_type = request.query_params['mediaOverviewType']
            today = datetime.datetime.today()
            datetime_recent = today - datetime.timedelta(days=365)

            if media_overview_type == MediaOverviewType.ALLTAGS.value:
                queryset = Timeline.objects.filter(
                    Q(year=today.year, month__lte=today.month)
                    | Q(year=datetime_recent.year, month__gte=datetime_recent.month)
                ).order_by("year", "month", "-article_count", "name")
            elif media_overview_type == MediaOverviewType.USERTAGS.value:
                tag_names = user_tags(request.query_params)
                queryset = (
                    Timeline.objects.filter(
                        Q(year=today.year, month__lte=today.month)
                        | Q(year=datetime_recent.year, month__gte=datetime_recent.month)
                    )
                    .filter(name__in=tag_names)
                    .order_by("year", "month", "-article_count", "name")
                )
            elif media_overview_type == MediaOverviewType.QUERYTAG.value:
                tag_names = [request.query_params['tag']]
                queryset = (
                    Timeline.objects.filter(
                        Q(year=today.year, month__lte=today.month)
                        | Q(year=datetime_recent.year, month__gte=datetime_recent.month)
                    )
                    .filter(name__in=tag_names)
                    .order_by("year", "month", "-article_count", "name")
                )
            serializer_class = TimelineSerializer(queryset, many=True)
            return JsonResponse(serializer_class.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


class OverviewList(APIView):
    """Contains all APIs regarding Overview objects"""

    today = datetime.datetime.today()
    datetime_recent = today - datetime.timedelta(days=90)

    @swagger_auto_schema(manual_parameters=[user_param_get], responses={200: OverviewSerializer(many=True)})
    def get(self, request):
        """
        Returns Overview objects for a users favorite tags in the last 90 days.

        :query_param user: id which is connected to the user (optional, default: 1)
        """
        try:
            if 'mediaOverviewType' in request.query_params:
                media_overview_type = request.query_params['mediaOverviewType']
            else:
                media_overview_type = MediaOverviewType.USERTAGS.value

            if media_overview_type == MediaOverviewType.ALLTAGS.value:
                overview_objects = Overview.objects.filter(
                    date_of_publication__range=[self.datetime_recent, self.today]
                )
            elif media_overview_type == MediaOverviewType.USERTAGS.value:
                tag_names = user_tags(request.query_params)
                overview_objects = Overview.objects.filter(
                    date_of_publication__range=[self.datetime_recent, self.today]
                ).filter(name__in=tag_names)
            overview_serializer = OverviewSerializer(overview_objects, many=True)
            return JsonResponse(overview_serializer.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


class Home(TemplateView):
    """Define the backends landing page that's shown when the base url is opened in a browser"""

    template_name = 'wizard/home.html'

    def get_context_data(self, **kwargs):
        """generate context data for Landing Page"""
        context = super().get_context_data(**kwargs)
        context['site_urls'] = {
            'articles': [
                {'name': 'open-api', 'url': '/swagger/'},
                {'name': 'articles', 'url': '/articles/articles'},
                {'name': 'media', 'url': '/articles/media'},
                {'name': 'medium type', 'url': '/articles/mediumtype'},
                {'name': 'by Tags', 'url': '/articles/byTags'},
                {'name': 'overview', 'url': '/articles/overview'},
            ],
            'pr_upload': [
                {'name': 'upload a file', 'url': '/pr_upload'},
            ],
            'train': [
                {'name': 'train data', 'url': '/prediction/train'},
            ],
            'wizard': [
                {'name': 'Wizard Overview', 'url': '/wizard/overview'},
            ],
            'projekt': [
                {'name': 'Repository', 'url': 'https://gitup.uni-potsdam.de/innohs/rest-api'},
                {
                    'name': 'Projektbeschreibung',
                    'url': 'https://www.uni-potsdam.de/de/multimedia/projekte/institutionelle-infrastrukturen/innovative-hochschule',
                },
            ],
        }
        return context


def create_function(itag):
    """
    Generate a graph, create a GenFunction object and create or update its FunctionData.

    :param itag: tag the generated graph is assigned to
    """
    date_bounds = (datetime.date(2021, 12, 1), datetime.date(2022, 6, 30))
    study_date = datetime.date(2022, 3, 1)
    study_follow_date = datetime.date(2022, 4, 1)
    date_delta = date_bounds[1] - date_bounds[0]
    all_x = np.array(range(0, date_delta.days))

    # creating function data
    x_data = [0, 0, 0, 0, 0]
    x_data[1] = (study_date - date_bounds[0]).days
    x_data[3] = (study_follow_date - date_bounds[0]).days
    x_data[2] = secrets.choice(list(range(x_data[1], x_data[3])))
    x_data[4] = x_data[1] + x_data[3]

    # randomly decide whether the curve should be rising, falling or stagnating
    randint = secrets.randbelow(3)
    if randint == 0:
        y_data = generate_rising_graph()
    elif randint == 1:
        y_data = generate_falling_graph()
    else:
        y_data = generate_stagnant_graph()

    series = poly.Polynomial.fit(x=x_data, y=y_data, deg=3)

    # creating y's of function
    y_data = []
    for part_x in all_x:
        value = series(part_x)
        value = secrets.choice([0, math.floor(value), math.ceil(value)])
        value = max(0, value)
        y_data.append(value)

    # writing function to database
    x_data = list(series(all_x))
    g_func = GenFunction(tag=itag, date_delta=date_delta.days)
    g_func.save()
    for data in list(zip(x_data, y_data)):
        f_data = FunctionData(x_data=data[0], y_data=data[1], tag=itag)
        f_data.save()


def generate_rising_graph():
    """Generate random y-values for a rising graph"""
    y_data = [0, 0, 0, 0, 0]
    y_data[0] = secrets.choice(list(range(1, 5)))
    y_data[1] = secrets.choice(list(range(5, 7)))
    y_data[2] = secrets.choice(list(range(2, 5)))
    y_data[3] = secrets.choice(list(range(6, 10)))
    y_data[4] = secrets.choice(list(range(8, 13)))
    return y_data


def generate_falling_graph():
    """Generate random y-values for a falling graph"""
    y_data = [0, 0, 0, 0, 0]
    y_data[0] = secrets.choice(list(range(8, 13)))
    y_data[1] = secrets.choice(list(range(6, 10)))
    y_data[2] = secrets.choice(list(range(2, 5)))
    y_data[3] = secrets.choice(list(range(5, 7)))
    y_data[4] = secrets.choice(list(range(1, 5)))
    return y_data


def generate_stagnant_graph():
    """Generate random y-values for a stagnant graph"""
    y_data = [0, 0, 0, 0, 0]
    y_data[0] = secrets.choice(list(range(2, 5)))
    y_data[1] = secrets.choice(list(range(2, 5)))
    y_data[2] = secrets.choice(list(range(2, 5)))
    y_data[3] = secrets.choice(list(range(2, 5)))
    y_data[4] = secrets.choice(list(range(2, 5)))
    return y_data


def generate_plot(data, date_delta):
    """
    Draw a plot from given function data

    :param data: list of tuples that contains (x,y,z) values
    :param date_delta: time period for which a plot is to be generated (difference between two datetime.date()'s)
    """
    x_date_range = pd.date_range('2021-12-01', periods=date_delta, freq='D')
    dframe = pd.DataFrame(
        data, index=x_date_range, columns=['scenario', 'generated article', 'accumulated gen. article/100']
    )
    fig = px.line(dframe)
    fig.update_traces(showlegend=False)
    return plot(fig, output_type='div')


def filter_by_regionality(media_regionality):
    """
    Return a list of media outlet names for a given regionality.

    :param media_regionality: MediaRegionality to filter by (e.g. MediaRegionality.REGIONAL)
    """
    outlets = []
    for media in Medium.objects.filter(medium_type__name=media_regionality.value).all():
        outlets.append(media.name)
    return outlets


def create_funcs(func):
    """
    Get plotting points from database and generate plot

    :param func: GenFunction object of the tag to create a plot for
    """
    qdata = FunctionData.objects.filter(tag=func.tag)
    x_data, y_data, z_data = [], [], []
    index = 0
    for qdat in qdata:
        index += qdat.y_data
        x_data.append(qdat.x_data)  # generated function
        y_data.append(qdat.y_data)  # aimed article generation per day
        z_data.append(index / 100)  # accumulated generated articles
    return generate_plot(list(zip(x_data, y_data, z_data)), func.date_delta)


class Logs(TemplateView):
    """Defines context and get methods for the 'Logs' wizard page"""

    template_name = 'wizard/logs.html'

    def get_context_data(self, **kwargs):
        """Get context data for logs page"""
        context = super().get_context_data(**kwargs)
        context['users'] = User.objects.all()

        logfiles = []
        for file in os.listdir(BASE_DIR):
            if file.startswith("oneup.log"):
                logfiles.append(file)

        context['logfiles'] = list(reversed(logfiles))

        if 'selectUser' in kwargs:
            context['user'] = 'user=' + str(kwargs['selectUser'])
            context['userid'] = str(kwargs['selectUser'])
        else:
            context['user'] = 'user=1'
            context['userid'] = '1'

        if 'selectFile' in kwargs:
            context['filename'] = str(kwargs['selectFile'])
        else:
            context['filename'] = 'oneup.log'

        with open(os.path.join(BASE_DIR, context['filename']), 'r', encoding='UTF-8') as log_file:
            context['logs'] = log_file.readlines()
            log_file.close()

        return context

    def get(self, request, *args, **kwargs):
        """Get the log page's template and context data"""
        if 'selectUser' in request.GET or 'selectFile' in request.GET:
            select_user_arg = request.GET['selectUser']
            if request.GET['selectUser'] == 'Select user id':
                select_user_arg = '1'

            select_file_arg = request.GET['selectFile']
            if request.GET['selectFile'] == 'Select log file':
                select_file_arg = 'oneup.log'

            context = self.get_context_data(selectUser=select_user_arg, selectFile=select_file_arg)
        else:
            context = self.get_context_data()
        return render(request, self.template_name, context)


class ArticleGenerator(TemplateView):
    """Defines context, post and get methods for the 'Article Generator' wizard page"""

    template_name = 'wizard/article_gen.html'

    def get_context_data(self, **kwargs):
        """Generate context data for article generation page"""
        context = super().get_context_data(**kwargs)

        if Tag.objects.count() == 0:
            generate_tag(0)
        l_graph_id = Tag.objects.all()[0].id
        if 'lGraph' in kwargs:
            l_graph_id = int(kwargs['lGraph'])
        tag_count = Tag.objects.count()

        context['tags'] = {
            'count': tag_count,
            'top': TopTags.objects.all().order_by("-count", "name"),
            'all': Tag.objects.exclude(name="account_used").all(),
        }
        context['wfunctions'] = {'all': GenFunction.objects.all()}

        # generate article creation functions for all tags
        if GenFunction.objects.count() < Tag.objects.count():
            for tag in Tag.objects.all():
                if len(GenFunction.objects.filter(tag=tag)) == 0:
                    create_function(tag)
        # FIXME: if GenFunction has no values in FunctionData, we maybe have to run another creation here

        # visualize function
        splot = [x for x in GenFunction.objects.all() if x.tag.id == l_graph_id][0]
        context['fplot'] = create_funcs(splot)
        context['lGraph'] = l_graph_id
        return context

    def post(self, request):
        """Generate graphs, articles or fake-tag depending on POST request."""
        if 'gGraph' in request.POST:
            FunctionData.objects.all().delete()
            GenFunction.objects.all().delete()
        elif 'gThisGraph' in request.POST:
            gtag = Tag.objects.all().filter(id=int(request.GET['lGraph']))[0]
            FunctionData.objects.filter(tag=gtag).delete()
            GenFunction.objects.filter(tag=gtag).delete()
        elif 'gArticle' in request.POST:
            generate_by_plot.delay()
        elif 'gThisArticle' in request.POST:
            generate_by_id.delay(int(request.GET['lGraph']))
        elif 'gTag' in request.POST:
            generate_tag(Tag.objects.count())
        return render(request, self.template_name, self.get_context_data())

    def get(self, request, *args, **kwargs):
        """
        Get the article generator page's template and context data
        and optionally select a specific graph to display on the page

        :http_param lGraph: raster_id of the tag which plot should be displayed (optional, default: 0)
        """
        logger.info('ArticleGenerator.get start')
        if 'lGraph' in request.GET:
            context = self.get_context_data(lGraph=request.GET['lGraph'])
        else:
            context = self.get_context_data()

        logger.info('ArticleGenerator.get ende')

        return render(request, self.template_name, context)


class OverviewDashboard(TemplateView):
    """Defines context and get method for the 'Overview' wizard page"""

    template_name = 'wizard/overview.html'

    def get_context_data(self, **kwargs):
        """Generate context data for overview page"""
        article_count = Article.objects.count()
        tag_count = Tag.objects.exclude(name="account_used").count()

        wrongly_parsed_articles = Article.objects.filter(wrongly_parsed=True).all()
        wrongly_parsed_files = Files.objects.filter(article__wrongly_parsed=True).all()

        context = super().get_context_data(**kwargs)
        context['articles'] = {
            'count': article_count,
            'wrongly_parsed_articles': wrongly_parsed_articles,
            'wrongly_parsed_files': wrongly_parsed_files,
            'wrongly_parsed_count': wrongly_parsed_articles.count(),
        }
        context['media'] = {
            'count': Medium.objects.count(),
            'top': TopMedia.objects.all(),
            'top_regional': [
                medium
                for medium in TopMedia.objects.all()
                if medium.name in filter_by_regionality(MediaRegionality.REGIONAL)
            ],
            'top_national': [
                medium
                for medium in TopMedia.objects.all()
                if medium.name in filter_by_regionality(MediaRegionality.NATIONAL)
            ],
            'top_website': [
                medium
                for medium in TopMedia.objects.all()
                if medium.name in filter_by_regionality(MediaRegionality.WEBSITE)
            ],
        }
        context['tags'] = {
            'count': tag_count,
            'top': TopTags.objects.all().order_by("-count", "name"),
            'all': Tag.objects.exclude(name="account_used").all(),
        }
        context['files'] = {'count': Files.objects.count()}

        available_users = User.objects.all()
        free_user_count = 0

        for user in available_users:
            if not user.favorite_tags.all():
                free_user_count += 1

        context['users'] = {
            'count': available_users.count(),
            'free_count': free_user_count
        }

        if (article_count == 0) or (tag_count == 0):
            context['ratio'] = {'apt': 0, 'tpa': 0}
        else:
            context['ratio'] = {'apt': article_count / tag_count, 'tpa': tag_count / article_count}
        return context

    def get(self, request, *args, **kwargs):
        """Get the overview page's template and context data"""
        context = self.get_context_data()
        return render(request, self.template_name, context)


class Simulator(View):
    """Defines post and get methods for the 'Simulator' wizard page"""

    template_name = 'wizard/simulator.html'
    selected_user = {}

    def get(self, request, *args, **kwargs):
        """Get the simulator page's template"""
        logger.debug("GET Simulator")
        return render(request, self.template_name, {})

    def post(self, request, *args, **kwargs):
        """On POST request, start parsing of all files in MEDIA_ROOT"""
        logger.info("POST Simulator")
        read_all_media_files.delay()
        autoremove_task.delay(MEDIA_ROOT)
        return render(request, self.template_name, {})


class UploadView(View):
    """*deprecated* Defines post and get methods for the 'pr_upload' file upload page"""

    template_name = 'articles/files_form.html'
    form_class = FileUploadForm

    def get(self, request, *args, **kwargs):
        """*deprecated* Get the file upload page's template and context data"""
        logger.debug("GET Uploadview")
        form = self.form_class(initial={'file_name': 'no_file.pdf'})
        return render(request, self.template_name, {'documents': Files.objects.all(), 'form': form})

    def post(self, request, *args, **kwargs):
        """*deprecated* Save the selected file (by file_name) to the database (if it's not saved already)"""
        logger.info("POST Uploadview")
        form = self.form_class(request.POST)
        if form.is_valid():
            name_ = form.cleaned_data['file_name']

            logger.debug("debug %i", Files.objects.filter(name=name_).count())
            if Files.objects.filter(name=name_).count() == 0:
                file = Files(name=name_, upload_date=timezone.now())
                file.save()
        return render(request, self.template_name, {'documents': Files.objects.all(), 'form': form})


class TopMediaView(APIView):
    """Return the media with the highest article count"""

    @swagger_auto_schema(manual_parameters=[user_param_get], responses={200: TopMediaSerializer(many=True)})
    def get(self, request):
        """
        Get the top media outlets for a specified media_overview_type
            - alltags: return top media for all existing tags
            - usertag: return top media for the users favorite tags
            - querytag: return top media only for the tag specified in the query

        :query_param mediaOverviewType: defines which tags should be taken into account (alltags, usertags, querytag)
        :query_param user: id which is connected to the user (optional, only considered when mediaOverviewType=usertags, default: 1)
        :query_param tag: name of the tag to return Timeline objects for (required when mediaOverviewType=querytag)
        """
        try:
            media_overview_type = request.query_params['mediaOverviewType']
            if media_overview_type == MediaOverviewType.ALLTAGS.value:
                # Counts the number of articles in a medium regardless of user tags
                top_media = (
                    Medium.objects.filter(article__wrongly_parsed=False)
                    .annotate(article_count=Count('article', distinct=True))
                    .order_by('-article_count', 'name')
                )
            elif media_overview_type == MediaOverviewType.USERTAGS.value:
                tag_names = user_tags(request.query_params)
                # Counts the number of articles in a medium depending on user tags
                top_media = (
                    Medium.objects.filter(article__tags__name__in=tag_names, article__wrongly_parsed=False)
                    .annotate(article_count=Count('article', distinct=True))
                    .order_by('-article_count', 'name')
                )
            elif media_overview_type == MediaOverviewType.QUERYTAG.value:
                tag_names = [request.query_params['tag']]
                # Counts the number of articles in a medium depending on query tag
                top_media = (
                    Medium.objects.filter(article__tags__name__in=tag_names, article__wrongly_parsed=False)
                    .annotate(article_count=Count('article', distinct=True))
                    .order_by('-article_count', 'name')
                )

            raw_media_data = []
            show_all_media = False
            if 'showAllMedia' in request.query_params:
                show_all_media = True
            for medium in top_media.all():
                raw_media_data.append({'medium_name': medium.name, 'article_count': medium.article_count})
                if not show_all_media and len(raw_media_data) > 4:
                    break
            top_media_serializer = TopMediaSerializer(raw_media_data, many=True)
            return JsonResponse(top_media_serializer.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


def unwanted_request(request, message):
    """If there are missing attributes in the request data, this function is taking care of logging"""
    logger.warning(message)
    logger.warning(request.data)
    return Response({'message': message}, 400)


user_param_post = openapi.Parameter(
    'user', openapi.IN_BODY, description="id which is connected to the user", type=openapi.TYPE_NUMBER
)

tags_param_post = openapi.Parameter(
    'tags',
    openapi.IN_BODY,
    description="array of tags",
    type=openapi.TYPE_ARRAY,
    items=openapi.Items(type=openapi.TYPE_STRING),
)


class UserFavoriteTags(APIView):
    """Defines get and post methods related to a users favorite tags"""

    @swagger_auto_schema(manual_parameters=[user_param_get], responses={200: TagSerializer(many=True)})
    def get(self, request):
        """
        Returns favorite tags for a specific user

        :query_param user: id which is connected to the user (optional, default: 1)
        """
        try:
            tag_names = user_tags(request.query_params)
            tags = Tag.objects.filter(name__in=tag_names, article__wrongly_parsed=False).distinct()
            tag_serializer = TagSerializer(tags, many=True)
            return JsonResponse(tag_serializer.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)

    # doesn't work yet
    @swagger_auto_schema(
        request_body=openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties={
                'user': openapi.Schema(type=openapi.TYPE_NUMBER, description='user_id'),
                'tags': openapi.Schema(
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Items(type=openapi.TYPE_STRING),
                    description='tags which should be saved',
                ),
            },
        ),
        responses={201: 'tags are set', 400: 'an parameter is wrongly set'},
    )
    def post(self, request):
        """
        Saves all given tags as a specific users favorite tags

        :request_data user: id which is connected to the user
        :request_data tags: array of tag names which should be set as the users favorite tags
        """
        if 'user' in request.data:
            if 'tags' in request.data:
                try:
                    user = User.objects.get(id_tag=request.data['user'])

                except ObjectDoesNotExist:
                    return unwanted_request(request, 'User not found')
                user.favorite_tags.remove(*user.favorite_tags.all())
                if isinstance(request.data['tags'], str):
                    sent_tags = request.data['tags'].split(', ')
                else:
                    sent_tags = request.data['tags']
                user.favorite_tags.add(*Tag.objects.filter(name__in=sent_tags).all())
                logger.info("Added tags.")
                logger.info(request.data)
                return Response({'message': 'Hello World!'}, 201)
            return unwanted_request(request, 'Couldn\'t update tags without tags')
        return unwanted_request(request, 'Couldn\'t update tags without user id')


def get_tag_count(tag_count, tags_dict, tag, att):
    """
    Get the article count of a specific tag if the tag name is in the list

    :param tag_count: object which holds tag_count for different time periods ('last_week', 'before_last_week', 'last_month', 'before_last_month')
    :param tags_dict: Tag object that holds all tags for the time period of interest
    :param tag: specific Tag object for which the article count should be retrieved
    :param att: name of the tag_count attribute which should be updated
    """
    filtered_week_days = list(filter(lambda t: t.name == tag.name, tags_dict.all()))  # search for tag.name in list
    if len(filtered_week_days) > 0:
        tag_count[att] = filtered_week_days[0].article_count


def get_filtered_tags(interval_date_start, interval_date_end):
    """
    Get tags sorted by article count in a specific time interval

    :param interval_date_start: datetime date which marks the start of the time interval
    :param interval_date_end: datetime date which marks the end of the time interval
    """
    return (
        Tag.objects.exclude(name="account_used")
        .filter(
            article__date_of_publication__range=[interval_date_start, interval_date_end - datetime.timedelta(days=1)],
            article__wrongly_parsed=False,
        )
        .annotate(article_count=Count('article', distinct=True))
        .order_by('-article_count')
    )


def create_dates(today):
    """
    This method creates a dict with dates of the last weeks, the last month and the month before the last month.
    The months are defined by the first and last day.

    :param today: datetime object that serves as a starting point for date calculations (e.g. datetime.datetime.today())
    """
    dates = {'today': today}
    dates['recent'] = dates['today'] - datetime.timedelta(days=60)
    dates['this_monday'] = dates['today'] - datetime.timedelta(days=dates['today'].weekday())
    dates['last_monday'] = dates['today'] - datetime.timedelta(days=dates['today'].weekday(), weeks=1)
    dates['before_last_monday'] = dates['today'] - datetime.timedelta(days=dates['today'].weekday(), weeks=2)
    dates['last_month'] = {}
    dates['before_last_month'] = {}
    dates['last_month']['last_day'] = dates['today'] - datetime.timedelta(days=dates['today'].day)
    dates['before_last_month']['last_day'] = dates['last_month']['last_day'] - datetime.timedelta(
        days=dates['last_month']['last_day'].day
    )
    dates['last_month']['first_day'] = dates['before_last_month']['last_day'] + datetime.timedelta(days=1)
    dates['before_last_month']['first_day'] = dates['before_last_month']['last_day'] - datetime.timedelta(
        days=dates['before_last_month']['last_day'].day - 1
    )
    return dates


def create_tag_diff_dict(dates):
    """
    Creates tags_dict object that holds tags for last_week, before_last_week, last_month, before_last_month

    :param dates: dates object created by create_dates()
    """
    tags_dict = {
        'last_week': get_filtered_tags(dates['last_monday'], dates['this_monday']),
        'before_last_week': get_filtered_tags(dates['before_last_monday'], dates['last_monday']),
        'last_month': get_filtered_tags(dates['last_month']['first_day'], dates['last_month']['last_day']),
        'before_last_month': get_filtered_tags(
            dates['before_last_month']['first_day'], dates['before_last_month']['last_day']
        ),
    }
    return tags_dict


class HotTopics(APIView):
    """
    This class calculates 'hot topics' by calculating which topics
    have been the most popular in a) the last week, b) the last month and
    how large the share of a topic in the totality of all topics is
    """

    @swagger_auto_schema(responses={200: TopTagsSerializer(many=True)})
    def get(self, request):
        """
        For every topic in a user's favorites, return the following data:

            - tag_name: name of the topic
            - article_count: total number of articles for that topic
            - hotness_factor: share of the topic's articles in the total number of all articles across all topics
            - diff_last_week: amount of topic's articles published in the last week
            - diff_last_month: amount of topic's articles published in the last month

        :query_param user: id which is connected to the user (optional, default: 1)
        """
        dates = create_dates(datetime.datetime.today())
        tags_dict = create_tag_diff_dict(dates)
        try:
            tag_names = user_tags(request.query_params)
            tags = (
                Tag.objects.exclude(name="account_used")
                .filter(
                    name__in=tag_names,
                    article__date_of_publication__range=[dates['recent'], dates['today']],
                    article__wrongly_parsed=False,
                )
                .annotate(article_count=Count('article', distinct=True))
                .order_by('-article_count', 'name')
            )

            tag_sum = tags.aggregate(Sum('article_count'))
            top_tags = []
            for tag in tags.all():
                tag_count = {'last_week': 0, 'before_last_week': 0, 'last_month': 0, 'before_last_month': 0}

                get_tag_count(tag_count, tags_dict['last_week'], tag, 'last_week')
                get_tag_count(tag_count, tags_dict['before_last_week'], tag, 'before_last_week')
                get_tag_count(tag_count, tags_dict['last_month'], tag, 'last_month')
                get_tag_count(tag_count, tags_dict['before_last_month'], tag, 'before_last_month')
                top_tags.append(
                    {
                        'tag_name': tag.name,
                        'article_count': tag.article_count,
                        'hotness_factor': tag.article_count / tag_sum['article_count__sum'],
                        'diff_last_week': tag_count['last_week'] - tag_count['before_last_week'],
                        'diff_last_month': tag_count['last_month'] - tag_count['before_last_month'],
                    }
                )
            top_tags_serializer = TopTagsSerializer(top_tags, many=True)
            return JsonResponse(top_tags_serializer.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


class TagsOverview(APIView):
    """Overview for tags grouped by parent tags"""

    queryset = Tag.objects.exclude(name="account_used").all()
    serializer_class = TagSerializer

    @swagger_auto_schema(responses={200: TagOverviewSerializer(many=True)})
    def get(self, request):
        """Returns all tags, grouped by parent tags"""
        parent_tags = ParentTag.objects.all()
        grouped_by_parent = []
        for parent_tag in parent_tags:
            tags = Tag.objects.filter(parent__name=parent_tag, article__wrongly_parsed=False).distinct()
            if tags:
                grouped_by_parent.append({'name': parent_tag.name, 'tags': tags})
        tag_overview_serializer = TagOverviewSerializer(grouped_by_parent, many=True)
        return JsonResponse(tag_overview_serializer.data, safe=False)


def sort_by_count(element):
    """Helper function to sort list by count"""
    return element['count']


class UserTagsArticleCount(APIView):
    """Class to see the influence of regional media on a topic"""

    @swagger_auto_schema(responses={200: TagsRegionalInfluenceSerializer(many=True)})
    def get(self, request):
        """
        Returns number of regional/national/web media per tag

        :query_param mediaOverviewType: defines which tags should be taken into account (alltags, usertags – optional, default: usertags)
        :query_param user: id which is connected to the user (optional, only considered when mediaOverviewType=usertags, default: 1)
        """
        try:
            if 'mediaOverviewType' in request.query_params:
                media_overview_type = request.query_params['mediaOverviewType']
            else:
                media_overview_type = MediaOverviewType.USERTAGS.value

            if media_overview_type == MediaOverviewType.ALLTAGS.value:
                tag_names = []
                for tag in Tag.objects.exclude(name="account_used").all():
                    tag_names.append(tag.name)
            elif media_overview_type == MediaOverviewType.USERTAGS.value:
                tag_names = user_tags(request.query_params)
            tag_regional_article_counts = []
            article_count_per_tag = (
                Medium.objects.filter(article__tags__name__in=tag_names)
                .values('article__tags__name')
                .annotate(article_count=Count('article', distinct=True))
            )
            article_count_per_tag_medium_type = (
                Medium.objects.filter(article__tags__name__in=tag_names)
                .values('article__tags__name', 'medium_type__name')
                .annotate(article_count=Count('article', distinct=True))
            )
            wrapper_dict = {}
            for tag in tag_names:
                wrapper_dict[tag] = {
                    MediaRegionality.REGIONAL.value: 0,
                    MediaRegionality.NATIONAL.value: 0,
                    MediaRegionality.WEBSITE.value: 0,
                    'article_count': 0,
                }
            for element in article_count_per_tag_medium_type:
                wrapper_dict[element['article__tags__name']][element['medium_type__name']] = element['article_count']
            for element in article_count_per_tag:
                wrapper_dict[element['article__tags__name']]['article_count'] = element['article_count']
            for key, value in wrapper_dict.items():
                tag_regional_article_counts.append(
                    {
                        'tag_name': key,
                        'count': value['article_count'],
                        'regional_count': value[MediaRegionality.REGIONAL.value],
                        'national_count': value[MediaRegionality.NATIONAL.value],
                        'web_count': value[MediaRegionality.WEBSITE.value],
                    }
                )
            tag_regional_article_counts.sort(reverse=True, key=sort_by_count)
            return JsonResponse(
                TagsRegionalInfluenceSerializer(tag_regional_article_counts, many=True).data, safe=False
            )
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


class RecentArticlesForMedium(APIView):
    """Get the most recent articles"""

    @swagger_auto_schema(responses={200: RecentArticleSerializer(many=True)})
    def get(self, request):
        """
        Returns an array of articles, sorted from most recent to oldest.

        :query_param mediaOverviewType: defines which tags should be taken into account (alltags, usertags, querytag)
        :query_param user: id which is connected to the user (optional, only considered when mediaOverviewType=usertags, default: 1)
        :query_param tag: name of the tag to return Timeline objects for (required when mediaOverviewType=querytag)
        :query_param medium_name: name of the medium for which recent articles should be returned ('all' for articles across all media)
        :query_param amountOfArticles: how many articles should be returned
        """
        try:
            media_overview_type = request.query_params['mediaOverviewType']
            medium = request.query_params['medium_name']
            amount_of_articles = int(request.query_params['amountOfArticles'])
            if media_overview_type == MediaOverviewType.ALLTAGS.value:
                if medium == 'all':
                    articles = (
                        Article.objects.filter(wrongly_parsed=False).all().order_by('-date_of_publication', 'name')
                    )
                else:
                    articles = (
                        Article.objects.filter(medium__name=medium, wrongly_parsed=False)
                        .all()
                        .order_by('-date_of_publication', 'name')
                    )
            else:
                if media_overview_type == MediaOverviewType.USERTAGS.value:
                    tag_names = user_tags(request.query_params)
                elif media_overview_type == MediaOverviewType.QUERYTAG.value:
                    tag_names = [request.query_params['tag']]

                if medium == 'all':
                    articles = (
                        Article.objects.filter(tags__name__in=tag_names, wrongly_parsed=False)
                        .all()
                        .order_by('-date_of_publication', 'name')
                    )
                else:
                    articles = (
                        Article.objects.filter(medium__name=medium, tags__name__in=tag_names, wrongly_parsed=False)
                        .all()
                        .order_by('-date_of_publication', 'name')
                    )

            recent_articles = []
            for article in articles.all():
                recent_articles.append(
                    {'name': article.name, 'date_of_publication': article.date_of_publication, 'text': article.text}
                )

                if len(recent_articles) > (amount_of_articles - 1):
                    break
            articles_serializer = RecentArticleSerializer(recent_articles, many=True)
            return JsonResponse(articles_serializer.data, safe=False)
        except ObjectDoesNotExist:
            logger.warning('User with id_tag %s does not exists', str(request.query_params['user']))
            return Response({'message': 'User does not exist.'}, 400)


class GetNewUserID(APIView):
    """Choose an unused user from all available users"""

    def get(self, request):
        """Requests an unused user id and instantly redirects the user to the Dashboard."""

        user_id_tag = get_free_user_id()

        if not user_id_tag:
            user_id_tag = '1'

        url = 'https://oneup-02.cs.uni-potsdam.de/dashboard/?user=' + user_id_tag

        return HttpResponseRedirect(url)

    @swagger_auto_schema(
        responses={
            200: openapi.Schema(
                type=openapi.TYPE_OBJECT, properties={'user-id': openapi.Schema(type=openapi.TYPE_NUMBER)}
            )
        }
    )
    def post(self, request):
        """Returns the next available, unused user id"""

        user_id_tag = get_free_user_id()

        if user_id_tag:
            return Response({"user-id": user_id_tag})

        # if this code is executed, no unused user was found
        logger.error("New user has been requested, but all users are already in use.")
        return Response({"user-id": "1", "error": "All users are already in use."})


def get_free_user_id():
    """Returns the next available, unused user id or 'None' if no user is available."""

    available_users = User.objects.all()

    for user in available_users:
        if not user.favorite_tags.all():
            logger.info("User " + str(user.id_tag) + " is unused and will be returned")

            try:
                tag_1 = Tag.objects.get(name="account_used")
            except ObjectDoesNotExist:
                tag_1 = Tag(name="account_used")
                tag_1.save()

            user.favorite_tags.add(tag_1)
            return str(user.id_tag)

    return None


class LogUserActivity(APIView):
    """This class provides an endpoint to log a users activity (f.e. idle, continue from idle)."""

    @swagger_auto_schema(
        responses={
            200: openapi.Schema(
                type=openapi.TYPE_OBJECT, properties={'message': openapi.Schema(type=openapi.TYPE_STRING)}
            )
        }
    )
    def post(self, request):
        """Accepts post requests. Since http-requests are logged automatically, this is empty for now."""

        return Response({"message": "executed LogUserActivity"})
